import grpc

import rainbow.protos.memory_pb2 as memory_pb2
import rainbow.protos.memory_pb2_grpc as memory_pb2_grpc

from .base import GraphBackend

# The memory database backend provides an interface to interact with an in memory cluster database


class MemoryBackend(GraphBackend):
    """
    A MemoryBackend is the rainbow default.

    This graph database backend is primarily for development.
    """

    def satisfies(self, jobspec):
        """
        Determine if a jobspec can be satisfied by the graph.
        """
        # Prepare a satisfy request with the jobspec
        # TODO if auth is in the graph, that needs to be done here too
        request = memory_pb2.SatisfyRequest(payload=jobspec.to_str())

        # Host should be set from the database_options from the client
        with grpc.insecure_channel(self.host) as channel:
            stub = memory_pb2_grpc.MemoryGraphStub(channel)
            response = stub.Satisfy(request)
        return response



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: memory.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0cmemory.proto\x12\x07service"C\n\x0fRegisterRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07payload\x18\x02 \x01(\t\x12\x11\n\tsubsystem\x18\x03 \x01(\t"!\n\x0eSatisfyRequest\x12\x0f\n\x07payload\x18\x01 \x01(\t"\xb3\x01\n\x0fSatisfyResponse\x12\x10\n\x08\x63lusters\x18\x01 \x03(\t\x12\x33\n\x06status\x18\x02 \x01(\x0e\x32#.service.SatisfyResponse.ResultType"Y\n\nResultType\x12\x1b\n\x17RESULT_TYPE_UNSPECIFIED\x10\x00\x12\x17\n\x13RESULT_TYPE_SUCCESS\x10\x01\x12\x15\n\x11RESULT_TYPE_ERROR\x10\x02"\x93\x01\n\x08Response\x12,\n\x06status\x18\x01 \x01(\x0e\x32\x1c.service.Response.ResultType"Y\n\nResultType\x12\x1b\n\x17RESULT_TYPE_UNSPECIFIED\x10\x00\x12\x17\n\x13RESULT_TYPE_SUCCESS\x10\x01\x12\x15\n\x11RESULT_TYPE_ERROR\x10\x02\x32\x88\x01\n\x0bMemoryGraph\x12>\n\x07Satisfy\x12\x17.service.SatisfyRequest\x1a\x18.service.SatisfyResponse"\x00\x12\x39\n\x08Register\x12\x18.service.RegisterRequest\x1a\x11.service.Response"\x00\x42@Z>github.com/converged-computing/rainbow/backends/memory/serviceb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "memory_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"Z>github.com/converged-computing/rainbow/backends/memory/service"
    _globals["_REGISTERREQUEST"]._serialized_start = 25
    _globals["_REGISTERREQUEST"]._serialized_end = 92
    _globals["_SATISFYREQUEST"]._serialized_start = 94
    _globals["_SATISFYREQUEST"]._serialized_end = 127
    _globals["_SATISFYRESPONSE"]._serialized_start = 130
    _globals["_SATISFYRESPONSE"]._serialized_end = 309
    _globals["_SATISFYRESPONSE_RESULTTYPE"]._serialized_start = 220
    _globals["_SATISFYRESPONSE_RESULTTYPE"]._serialized_end = 309
    _globals["_RESPONSE"]._serialized_start = 312
    _globals["_RESPONSE"]._serialized_end = 459
    _globals["_RESPONSE_RESULTTYPE"]._serialized_start = 220
    _globals["_RESPONSE_RESULTTYPE"]._serialized_end = 309
    _globals["_MEMORYGRAPH"]._serialized_start = 462
    _globals["_MEMORYGRAPH"]._serialized_end = 598
# @@protoc_insertion_point(module_scope)



import logging
from concurrent import futures

import grpc

from rainbow.protos import api_pb2_grpc


class RainbowSchedulerServicer(api_pb2_grpc.RainbowSchedulerServicer):
    """
    Unimplemented server - let us know if you need this, but largely
    you should use Go for the serve and python for a client only.
    """

    pass


def serve():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    api_pb2_grpc.add_RainbowSchedulerServicer_to_server(RainbowSchedulerServicer(), server)
    server.add_insecure_port("[::]:50051")
    server.start()
    server.wait_for_termination()


if __name__ == "__main__":
    logging.basicConfig()
    serve()



import json

import yaml


def read_json(filename):
    """
    Read json from file
    """
    return json.loads(read_file(filename))


def read_file(filename):
    """
    Read in a file content
    """
    with open(filename, "r") as fd:
        content = fd.read()
    return content


def read_yaml(filename):
    """
    Read yaml from file
    """
    with open(filename, "r") as fd:
        content = yaml.safe_load(fd)
    return content


def write_yaml(obj, filename):
    """
    Read yaml to file
    """
    with open(filename, "w") as fd:
        yaml.dump(obj, fd)



import os

import jsonschema

import rainbow.schema as schemas
import rainbow.utils as utils


def new_rainbow_config(host, cluster, secret, scheduler_name="rainbow"):
    template = {
        "scheduler": {
            "name": scheduler_name,
            "secret": secret,
            "algorithm": {"name": "random"},
        },
        "cluster": {
            "name": cluster,
        },
        "graphdatabase": {
            "name": "memory",
            "options": {"host": host},
        },
    }
    cfg = RainbowConfig()
    cfg._cfg = template
    cfg.validate()
    return cfg


class RainbowConfig:
    """
    A RainbowClient is able to interact with a Rainbow cluster from Python.
    """

    def __init__(self, config_file=None, validate=True):
        """
        Create a new rainbow client to interact with a rainbow cluster.
        """
        self.config_file = config_file
        self._cfg = None
        if self.config_file and os.path.exists(self.config_file):
            self.load()
            if validate:
                self.validate()

    @property
    def cfg(self):
        """
        Helper function to get/load the config if not done yet.
        """
        if self._cfg is None:
            self.load()
        return self._cfg

    def load(self, config_file=None):
        """
        Load a rainbow config
        """
        config_file = config_file or self.config_file

        # The config is required
        if not config_file or not os.path.exists(config_file):
            raise ValueError("This functionality requires a <instance>.config_file")
        self._cfg = utils.read_yaml(config_file)
        self.config_file = config_file
        return self._cfg

    def validate(self):
        """
        Validate the config against its schema
        """
        if not self._cfg:
            return
        jsonschema.validate(self._cfg, schema=schemas.rainbow_config_v1)

    def save_yaml(self, path):
        """
        Write yaml to file
        """
        utils.write_yaml(self._cfg, path)

    def remove_cluster(self, name):
        """
        Remove a cluster from the listing.
        """
        updated = []
        for cluster in self._cfg["clusters"]:
            if cluster["name"] == name:
                continue
            updated.append(cluster)
        self._cfg["clusters"] = updated

    def add_cluster(self, name, token):
        """
        Add a cluster to the listing.
        """
        if not self._cfg:
            return

        if "clusters" not in self._cfg:
            self._cfg["clusters"] = []

        # Ensure we don't have the cluster already
        for cluster in self._cfg["clusters"]:
            if cluster["name"] == name:
                raise ValueError(f"cluster {name} already exists - remove it first")
        self._cfg["clusters"].append({"name": name, "token": token})

    def get_clusters(self, names):
        """
        Get clusters, optionally filtering to a set of names
        """
        clusters = []
        if not self._cfg:
            return clusters

        for cluster in self._cfg.get("clusters", []):
            name = cluster.get("name")
            token = cluster.get("token")
            if not name or not token:
                continue

            # Are we filtering down to a set?
            if names and name not in names:
                continue
            clusters.append(cluster)
        return clusters

    def get_database(self):
        """
        Get the database defined in the config, if defined
        """
        # Cut out early if we don't have a config
        if self._cfg is None:
            return
        return self.cfg.get("graphdatabase")



# A graph database backend provides the basic functions to interact with
# the graph database of rainbow, from a client perspective


class GraphBackend:
    """
    The GraphBackend is the base backend that primarily
    shows the abstract functions to be defined. Yes I'm not
    using abc. This is yet-another-hill... :)
    """

    def __init__(self, **options):
        """
        Create a new graph backend, accepting any options type.
        """
        # Set options as attributes
        for key, value in options.items():
            setattr(self, key, value)

    def satisfies(self, jobspec):
        """
        Determine if a jobspec can be satisfied
        """
        raise NotImplementedError



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import rainbow_pb2 as rainbow__pb2


class RainbowSchedulerStub(object):
    """RainbowSchedulerService provides API endpoints for interacting with the central scheduler service"""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Register = channel.unary_unary(
            "/convergedcomputing.org.grpc.v1.RainbowScheduler/Register",
            request_serializer=rainbow__pb2.RegisterRequest.SerializeToString,
            response_deserializer=rainbow__pb2.RegisterResponse.FromString,
        )
        self.RegisterSubsystem = channel.unary_unary(
            "/convergedcomputing.org.grpc.v1.RainbowScheduler/RegisterSubsystem",
            request_serializer=rainbow__pb2.RegisterRequest.SerializeToString,
            response_deserializer=rainbow__pb2.RegisterResponse.FromString,
        )
        self.SubmitJob = channel.unary_unary(
            "/convergedcomputing.org.grpc.v1.RainbowScheduler/SubmitJob",
            request_serializer=rainbow__pb2.SubmitJobRequest.SerializeToString,
            response_deserializer=rainbow__pb2.SubmitJobResponse.FromString,
        )
        self.ReceiveJobs = channel.unary_unary(
            "/convergedcomputing.org.grpc.v1.RainbowScheduler/ReceiveJobs",
            request_serializer=rainbow__pb2.ReceiveJobsRequest.SerializeToString,
            response_deserializer=rainbow__pb2.ReceiveJobsResponse.FromString,
        )
        self.AcceptJobs = channel.unary_unary(
            "/convergedcomputing.org.grpc.v1.RainbowScheduler/AcceptJobs",
            request_serializer=rainbow__pb2.AcceptJobsRequest.SerializeToString,
            response_deserializer=rainbow__pb2.AcceptJobsResponse.FromString,
        )


class RainbowSchedulerServicer(object):
    """RainbowSchedulerService provides API endpoints for interacting with the central scheduler service"""

    def Register(self, request, context):
        """Register cluster - request to register a new cluster"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RegisterSubsystem(self, request, context):
        """Register cluster - request to register a new cluster"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SubmitJob(self, request, context):
        """Job Submission - request for submitting a job to a named cluster"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ReceiveJobs(self, request, context):
        """Request Job - ask the rainbow scheduler for up to max jobs"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def AcceptJobs(self, request, context):
        """Accept Jobs - accept some number of jobs"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_RainbowSchedulerServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Register": grpc.unary_unary_rpc_method_handler(
            servicer.Register,
            request_deserializer=rainbow__pb2.RegisterRequest.FromString,
            response_serializer=rainbow__pb2.RegisterResponse.SerializeToString,
        ),
        "RegisterSubsystem": grpc.unary_unary_rpc_method_handler(
            servicer.RegisterSubsystem,
            request_deserializer=rainbow__pb2.RegisterRequest.FromString,
            response_serializer=rainbow__pb2.RegisterResponse.SerializeToString,
        ),
        "SubmitJob": grpc.unary_unary_rpc_method_handler(
            servicer.SubmitJob,
            request_deserializer=rainbow__pb2.SubmitJobRequest.FromString,
            response_serializer=rainbow__pb2.SubmitJobResponse.SerializeToString,
        ),
        "ReceiveJobs": grpc.unary_unary_rpc_method_handler(
            servicer.ReceiveJobs,
            request_deserializer=rainbow__pb2.ReceiveJobsRequest.FromString,
            response_serializer=rainbow__pb2.ReceiveJobsResponse.SerializeToString,
        ),
        "AcceptJobs": grpc.unary_unary_rpc_method_handler(
            servicer.AcceptJobs,
            request_deserializer=rainbow__pb2.AcceptJobsRequest.FromString,
            response_serializer=rainbow__pb2.AcceptJobsResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "convergedcomputing.org.grpc.v1.RainbowScheduler", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class RainbowScheduler(object):
    """RainbowSchedulerService provides API endpoints for interacting with the central scheduler service"""

    @staticmethod
    def Register(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/convergedcomputing.org.grpc.v1.RainbowScheduler/Register",
            rainbow__pb2.RegisterRequest.SerializeToString,
            rainbow__pb2.RegisterResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RegisterSubsystem(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/convergedcomputing.org.grpc.v1.RainbowScheduler/RegisterSubsystem",
            rainbow__pb2.RegisterRequest.SerializeToString,
            rainbow__pb2.RegisterResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SubmitJob(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/convergedcomputing.org.grpc.v1.RainbowScheduler/SubmitJob",
            rainbow__pb2.SubmitJobRequest.SerializeToString,
            rainbow__pb2.SubmitJobResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ReceiveJobs(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/convergedcomputing.org.grpc.v1.RainbowScheduler/ReceiveJobs",
            rainbow__pb2.ReceiveJobsRequest.SerializeToString,
            rainbow__pb2.ReceiveJobsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def AcceptJobs(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/convergedcomputing.org.grpc.v1.RainbowScheduler/AcceptJobs",
            rainbow__pb2.AcceptJobsRequest.SerializeToString,
            rainbow__pb2.AcceptJobsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



import json
import os

import grpc

import rainbow.backends as backends
import rainbow.config as config
import rainbow.defaults as defaults
import rainbow.jobspec as js
import rainbow.jobspec.converter as converter
import rainbow.utils as utils
from rainbow.protos import rainbow_pb2, rainbow_pb2_grpc

# TODO need to register the databases here...


class RainbowClient:
    """
    A RainbowClient is able to interact with a Rainbow cluster from Python.
    """

    def __init__(self, host="localhost:50051", config_file=None, database=None):
        """
        Create a new rainbow client to interact with a rainbow cluster.
        """
        self.cfg = config.RainbowConfig(config_file)
        self.host = host or self.cfg.get("scheduler", {}).get("host")

        # load the graph database backend
        self.set_database(database)
        self.load_backend()

    def receive_jobs(self, max_jobs=None):
        """
        Receive jobs (query and send accept response back to rainbow)
        """
        jobs = []

        # This requires a cluster to be defined
        cfg = self.cfg
        cluster = cfg._cfg.get("cluster")
        if "name" not in cluster or "secret" not in cluster:
            raise ValueError("'cluster' defined in the rainbow config needs a name and secret")

        # These are submit variables. A more substantial submit script would have argparse, etc.
        request = rainbow_pb2.ReceiveJobsRequest(secret=cluster["secret"], cluster=cluster["name"])
        # This defaults to 0, so only set of non 0 or not None
        if max_jobs:
            request.maxJobs = max_jobs

        with grpc.insecure_channel(self.host) as channel:
            stub = rainbow_pb2_grpc.RainbowSchedulerStub(channel)
            response = stub.ReceiveJobs(request)

            # Case 1: no jobs to receive
            if response.status == 0:
                print("There are no jobs to receive.")
                return jobs

            if response.status != 1:
                print("Issue with requesting jobs:")
                return jobs

        print("Status: REQUEST_JOBS_SUCCESS")
        print(f"Received {len(response.jobs)} jobs to accept...")
        jobs = [json.loads(job) for job in list(response.jobs.values())]

        # Tell rainbow we accepted them
        jobids = list(response.jobs.keys())
        request = rainbow_pb2.AcceptJobsRequest(
            cluster=cluster["name"],
            secret=cluster["secret"],
            jobids=jobids,
        )
        with grpc.insecure_channel(self.host) as channel:
            stub = rainbow_pb2_grpc.RainbowSchedulerStub(channel)
            response = stub.AcceptJobs(request)
        return jobs

    def register(self, cluster, secret, cluster_nodes):
        """
        Register a cluster to the Rainbow Scheduler.
        """
        if not cluster:
            raise ValueError("A cluster name is required to register")
        if not secret:
            raise ValueError("A secret is required to register")
        if not os.path.exists(cluster_nodes):
            raise ValueError(f"Cluster nodes file {cluster_nodes} for registration does not exist.")

        # Read in the nodes to string
        nodes = utils.read_file(cluster_nodes)

        # These are the variables for our cluster - name for now
        registerRequest = rainbow_pb2.RegisterRequest(
            name=cluster,
            secret=secret,
            nodes=nodes,
        )

        with grpc.insecure_channel(self.host) as channel:
            stub = rainbow_pb2_grpc.RainbowSchedulerStub(channel)
            response = stub.Register(registerRequest)
        return response

    def register_subsystem(self, cluster, subsystem, secret, nodes):
        """
        Register subsystem nodes to rainbow
        """
        if not cluster:
            raise ValueError("A cluster name is required to register")
        if not secret:
            raise ValueError("A secret is required to register")
        if not os.path.exists(nodes):
            raise ValueError(f"Subsystem nodes file {nodes} for registration does not exist.")

        # Read in the nodes to string
        nodes = utils.read_file(nodes)

        # These are the variables for our cluster - name for now
        registerRequest = rainbow_pb2.RegisterRequest(
            name=cluster,
            secret=secret,
            nodes=nodes,
            subsystem=subsystem,
        )

        with grpc.insecure_channel(self.host) as channel:
            stub = rainbow_pb2_grpc.RainbowSchedulerStub(channel)
            response = stub.RegisterSubsystem(registerRequest)
        return response

    def load_backend(self):
        """
        Load the backend
        """
        # This will error if you do something wonky
        self.backend = backends.get_backend(self.database, self.database_options)

    def set_database(self, database=None):
        """
        Load the graph database backend, and options
        """
        # It's not technically required, so touch lightly
        # Only go through this if a database is not provided
        options = {"host": defaults.database_host}
        db = self.cfg.get_database()
        if db is not None:
            database = db.get("name")
            options = db.get("options") or options

        # If we get here and no database, use the default
        if not database:
            database = defaults.database_backend
        self.database = database
        self.database_options = options

    def submit_jobspec(self, jobspec):
        """
        Submit a jobspec directly. This is useful if you want to generate
        it custom with your own special logic.
        """
        # Ask the database backend if our jobspec can be satisfied
        response = self.backend.satisfies(jobspec)
        matches = response.clusters
        print(response)

        # No matches?
        if not matches:
            print("No clusters match the request")
            return response

        # TODO these need to have (again) the token and name checked
        # This is backwards because we check the token AFTER getting it, and it needs
        # to go to the graph (request above). I haven't implemented this yet.
        matches = self.cfg.get_clusters(matches)
        clusters = []
        for match in matches:
            clusters.append(
                rainbow_pb2.SubmitJobRequest.Cluster(name=match["name"], token=match["token"])
            )

        # THEN contact rainbwo with clusters
        # These are submit variables. A more substantial submit script would have argparse, etc.
        submitRequest = rainbow_pb2.SubmitJobRequest(
            name=jobspec.name,
            clusters=clusters,
            jobspec=jobspec.to_yaml(),
        )

        with grpc.insecure_channel(self.host) as channel:
            stub = rainbow_pb2_grpc.RainbowSchedulerStub(channel)
            response = stub.SubmitJob(submitRequest)
        return response

    def submit_job(self, command, nodes=1, tasks=1):
        """
        Submit a simple job to rainbow. This includes:

        1. Converting the basic attributes into a jobspec (and validating)
        2. Reading and validating the cluster config
        3. Asking the graph for which clusters satisfy
        4. Then submitting to rainbow
        """
        if not command:
            raise ValueError("A command is required")

        # Pretty print for the user (arguably we don't need this, oh well)
        cmd = command
        if isinstance(cmd, list):
            command = " ".join(cmd)
            print(f"⭐️ Submitting job: {cmd}")

        # Generate the jobspec dictionary
        raw = converter.new_simple_jobspec(nodes=nodes, command=command, tasks=tasks)
        jobspec = js.Jobspec(raw)
        return self.submit_jobspec(jobspec)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import memory_pb2 as memory__pb2


class MemoryGraphStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Satisfy = channel.unary_unary(
            "/service.MemoryGraph/Satisfy",
            request_serializer=memory__pb2.SatisfyRequest.SerializeToString,
            response_deserializer=memory__pb2.SatisfyResponse.FromString,
        )
        self.Register = channel.unary_unary(
            "/service.MemoryGraph/Register",
            request_serializer=memory__pb2.RegisterRequest.SerializeToString,
            response_deserializer=memory__pb2.Response.FromString,
        )


class MemoryGraphServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Satisfy(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Register(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_MemoryGraphServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Satisfy": grpc.unary_unary_rpc_method_handler(
            servicer.Satisfy,
            request_deserializer=memory__pb2.SatisfyRequest.FromString,
            response_serializer=memory__pb2.SatisfyResponse.SerializeToString,
        ),
        "Register": grpc.unary_unary_rpc_method_handler(
            servicer.Register,
            request_deserializer=memory__pb2.RegisterRequest.FromString,
            response_serializer=memory__pb2.Response.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "service.MemoryGraph", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class MemoryGraph(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Satisfy(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/service.MemoryGraph/Satisfy",
            memory__pb2.SatisfyRequest.SerializeToString,
            memory__pb2.SatisfyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Register(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/service.MemoryGraph/Register",
            memory__pb2.RegisterRequest.SerializeToString,
            memory__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class RegisterRequest(_message.Message):
    __slots__ = ("name", "payload", "subsystem")
    NAME_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    SUBSYSTEM_FIELD_NUMBER: _ClassVar[int]
    name: str
    payload: str
    subsystem: str
    def __init__(self, name: _Optional[str] = ..., payload: _Optional[str] = ..., subsystem: _Optional[str] = ...) -> None: ...

class SatisfyRequest(_message.Message):
    __slots__ = ("payload",)
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    payload: str
    def __init__(self, payload: _Optional[str] = ...) -> None: ...

class SatisfyResponse(_message.Message):
    __slots__ = ("clusters", "status")
    class ResultType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        RESULT_TYPE_UNSPECIFIED: _ClassVar[SatisfyResponse.ResultType]
        RESULT_TYPE_SUCCESS: _ClassVar[SatisfyResponse.ResultType]
        RESULT_TYPE_ERROR: _ClassVar[SatisfyResponse.ResultType]
    RESULT_TYPE_UNSPECIFIED: SatisfyResponse.ResultType
    RESULT_TYPE_SUCCESS: SatisfyResponse.ResultType
    RESULT_TYPE_ERROR: SatisfyResponse.ResultType
    CLUSTERS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    clusters: _containers.RepeatedScalarFieldContainer[str]
    status: SatisfyResponse.ResultType
    def __init__(self, clusters: _Optional[_Iterable[str]] = ..., status: _Optional[_Union[SatisfyResponse.ResultType, str]] = ...) -> None: ...

class Response(_message.Message):
    __slots__ = ("status",)
    class ResultType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        RESULT_TYPE_UNSPECIFIED: _ClassVar[Response.ResultType]
        RESULT_TYPE_SUCCESS: _ClassVar[Response.ResultType]
        RESULT_TYPE_ERROR: _ClassVar[Response.ResultType]
    RESULT_TYPE_UNSPECIFIED: Response.ResultType
    RESULT_TYPE_SUCCESS: Response.ResultType
    RESULT_TYPE_ERROR: Response.ResultType
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: Response.ResultType
    def __init__(self, status: _Optional[_Union[Response.ResultType, str]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rainbow.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rrainbow.proto\x12\x1e\x63onvergedcomputing.org.grpc.v1\x1a\x1fgoogle/protobuf/timestamp.proto"{\n\x0fRegisterRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06secret\x18\x02 \x01(\t\x12\r\n\x05nodes\x18\x03 \x01(\t\x12\x11\n\tsubsystem\x18\x04 \x01(\t\x12(\n\x04sent\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp"\xcf\x01\n\x10SubmitJobRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12J\n\x08\x63lusters\x18\x02 \x03(\x0b\x32\x38.convergedcomputing.org.grpc.v1.SubmitJobRequest.Cluster\x12\x0f\n\x07jobspec\x18\x03 \x01(\t\x12(\n\x04sent\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a&\n\x07\x43luster\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05token\x18\x02 \x01(\t"p\n\x12ReceiveJobsRequest\x12\x0f\n\x07\x63luster\x18\x01 \x01(\t\x12\x0e\n\x06secret\x18\x02 \x01(\t\x12\x0f\n\x07maxJobs\x18\x03 \x01(\x05\x12(\n\x04sent\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp"n\n\x11\x41\x63\x63\x65ptJobsRequest\x12\x0f\n\x07\x63luster\x18\x01 \x01(\t\x12\x0e\n\x06secret\x18\x02 \x01(\t\x12\x0e\n\x06jobids\x18\x03 \x03(\x05\x12(\n\x04sent\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp"\x8e\x02\n\x10RegisterResponse\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\r\n\x05token\x18\x02 \x01(\t\x12\x0e\n\x06secret\x18\x03 \x01(\t\x12K\n\x06status\x18\x04 \x01(\x0e\x32;.convergedcomputing.org.grpc.v1.RegisterResponse.ResultType"z\n\nResultType\x12\x18\n\x14REGISTER_UNSPECIFIED\x10\x00\x12\x14\n\x10REGISTER_SUCCESS\x10\x01\x12\x12\n\x0eREGISTER_ERROR\x10\x02\x12\x13\n\x0fREGISTER_DENIED\x10\x03\x12\x13\n\x0fREGISTER_EXISTS\x10\x04"\xe3\x01\n\x11SubmitJobResponse\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\r\n\x05jobid\x18\x02 \x01(\x05\x12L\n\x06status\x18\x03 \x01(\x0e\x32<.convergedcomputing.org.grpc.v1.SubmitJobResponse.ResultType"]\n\nResultType\x12\x16\n\x12SUBMIT_UNSPECIFIED\x10\x00\x12\x12\n\x0eSUBMIT_SUCCESS\x10\x01\x12\x10\n\x0cSUBMIT_ERROR\x10\x02\x12\x11\n\rSUBMIT_DENIED\x10\x03"\xe8\x02\n\x13ReceiveJobsResponse\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12K\n\x04jobs\x18\x02 \x03(\x0b\x32=.convergedcomputing.org.grpc.v1.ReceiveJobsResponse.JobsEntry\x12N\n\x06status\x18\x03 \x01(\x0e\x32>.convergedcomputing.org.grpc.v1.ReceiveJobsResponse.ResultType\x1a+\n\tJobsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"s\n\nResultType\x12\x1a\n\x16REQUEST_JOBS_NORESULTS\x10\x00\x12\x18\n\x14REQUEST_JOBS_SUCCESS\x10\x01\x12\x16\n\x12REQUEST_JOBS_ERROR\x10\x02\x12\x17\n\x13REQUEST_JOBS_DENIED\x10\x03"\xd7\x01\n\x12\x41\x63\x63\x65ptJobsResponse\x12M\n\x06status\x18\x01 \x01(\x0e\x32=.convergedcomputing.org.grpc.v1.AcceptJobsResponse.ResultType"r\n\nResultType\x12\x1b\n\x17RESULT_TYPE_UNSPECIFIED\x10\x00\x12\x17\n\x13RESULT_TYPE_PARTIAL\x10\x01\x12\x17\n\x13RESULT_TYPE_SUCCESS\x10\x02\x12\x15\n\x11RESULT_TYPE_ERROR\x10\x03\x32\xd8\x04\n\x10RainbowScheduler\x12m\n\x08Register\x12/.convergedcomputing.org.grpc.v1.RegisterRequest\x1a\x30.convergedcomputing.org.grpc.v1.RegisterResponse\x12v\n\x11RegisterSubsystem\x12/.convergedcomputing.org.grpc.v1.RegisterRequest\x1a\x30.convergedcomputing.org.grpc.v1.RegisterResponse\x12p\n\tSubmitJob\x12\x30.convergedcomputing.org.grpc.v1.SubmitJobRequest\x1a\x31.convergedcomputing.org.grpc.v1.SubmitJobResponse\x12v\n\x0bReceiveJobs\x12\x32.convergedcomputing.org.grpc.v1.ReceiveJobsRequest\x1a\x33.convergedcomputing.org.grpc.v1.ReceiveJobsResponse\x12s\n\nAcceptJobs\x12\x31.convergedcomputing.org.grpc.v1.AcceptJobsRequest\x1a\x32.convergedcomputing.org.grpc.v1.AcceptJobsResponseB3Z1github.com/converged-computing/rainbow/pkg/api/v1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "rainbow_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"Z1github.com/converged-computing/rainbow/pkg/api/v1"
    _globals["_RECEIVEJOBSRESPONSE_JOBSENTRY"]._options = None
    _globals["_RECEIVEJOBSRESPONSE_JOBSENTRY"]._serialized_options = b"8\001"
    _globals["_REGISTERREQUEST"]._serialized_start = 82
    _globals["_REGISTERREQUEST"]._serialized_end = 205
    _globals["_SUBMITJOBREQUEST"]._serialized_start = 208
    _globals["_SUBMITJOBREQUEST"]._serialized_end = 415
    _globals["_SUBMITJOBREQUEST_CLUSTER"]._serialized_start = 377
    _globals["_SUBMITJOBREQUEST_CLUSTER"]._serialized_end = 415
    _globals["_RECEIVEJOBSREQUEST"]._serialized_start = 417
    _globals["_RECEIVEJOBSREQUEST"]._serialized_end = 529
    _globals["_ACCEPTJOBSREQUEST"]._serialized_start = 531
    _globals["_ACCEPTJOBSREQUEST"]._serialized_end = 641
    _globals["_REGISTERRESPONSE"]._serialized_start = 644
    _globals["_REGISTERRESPONSE"]._serialized_end = 914
    _globals["_REGISTERRESPONSE_RESULTTYPE"]._serialized_start = 792
    _globals["_REGISTERRESPONSE_RESULTTYPE"]._serialized_end = 914
    _globals["_SUBMITJOBRESPONSE"]._serialized_start = 917
    _globals["_SUBMITJOBRESPONSE"]._serialized_end = 1144
    _globals["_SUBMITJOBRESPONSE_RESULTTYPE"]._serialized_start = 1051
    _globals["_SUBMITJOBRESPONSE_RESULTTYPE"]._serialized_end = 1144
    _globals["_RECEIVEJOBSRESPONSE"]._serialized_start = 1147
    _globals["_RECEIVEJOBSRESPONSE"]._serialized_end = 1507
    _globals["_RECEIVEJOBSRESPONSE_JOBSENTRY"]._serialized_start = 1347
    _globals["_RECEIVEJOBSRESPONSE_JOBSENTRY"]._serialized_end = 1390
    _globals["_RECEIVEJOBSRESPONSE_RESULTTYPE"]._serialized_start = 1392
    _globals["_RECEIVEJOBSRESPONSE_RESULTTYPE"]._serialized_end = 1507
    _globals["_ACCEPTJOBSRESPONSE"]._serialized_start = 1510
    _globals["_ACCEPTJOBSRESPONSE"]._serialized_end = 1725
    _globals["_ACCEPTJOBSRESPONSE_RESULTTYPE"]._serialized_start = 1611
    _globals["_ACCEPTJOBSRESPONSE_RESULTTYPE"]._serialized_end = 1725
    _globals["_RAINBOWSCHEDULER"]._serialized_start = 1728
    _globals["_RAINBOWSCHEDULER"]._serialized_end = 2328
# @@protoc_insertion_point(module_scope)



import shlex


def new_simple_jobspec(command, nodes=1, name=None, tasks=1, jobspec_version=1):
    """
    Generate a new simple jobspec from basic parameters.
    """
    if isinstance(command, str):
        command = shlex.split(command)

    if not command:
        raise ValueError("A command must be provided.")

    # If we don't have a name, derive one
    if name is None:
        name = command[0]

    if nodes < 1 or tasks < 1:
        raise ValueError("Nodes and tasks for the job must be >= 1")

    node_resource = {
        "type": "node",
        "count": nodes,
    }
    slot = {
        "type": "slot",
        "count": 1,
        "label": name,
    }
    task_resource = {
        "type": "core",
        "count": tasks,
    }
    slot["with"] = [task_resource]

    node_resource["with"] = [slot]
    task_resources = [
        {
            "command": command,
            "slot": name,
            "count": {
                "per_slot": 1,
            },
        }
    ]
    return {
        "version": jobspec_version,
        "resources": [node_resource],
        "tasks": task_resources,
        "attributes": {},
    }



import os

from setuptools import find_packages, setup  # noqa: H301

DESCRIPTION = "Python gRPC functions for the Rainbow Scheduler"

# Try to read description, otherwise fallback to short description
try:
    with open(os.path.abspath("README.md")) as filey:
        LONG_DESCRIPTION = filey.read()
except Exception:
    LONG_DESCRIPTION = DESCRIPTION

################################################################################
# MAIN #########################################################################
################################################################################

if __name__ == "__main__":
    setup(
        name="rainbow-scheduler",
        version="0.0.12",
        author="Vanessasaurus",
        author_email="vsoch@users.noreply.github.com",
        maintainer="Vanessasaurus",
        packages=find_packages(),
        include_package_data=True,
        zip_safe=False,
        url="https://github.com/converged-computing/rainbow/tree/main/python/v1",
        license="MIT",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        keywords="multi-cluster, scheduler",
        setup_requires=["pytest-runner"],
        install_requires=["grpcio", "grpcio-tools", "jsonschema", "pyyaml"],
        tests_require=["pytest", "pytest-cov"],
        classifiers=[
            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: C",
            "Programming Language :: Python",
            "Topic :: Software Development",
            "Topic :: Scientific/Engineering",
            "Operating System :: Unix",
            "Programming Language :: Python :: 3.7",
        ],
    )



rainbow_config_v1 = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "http://github.com/flux-framework/rfc/tree/master/data/spec_24/schema.json",
    "title": "rainbow-config-01",
    "description": "Rainbow config version 1.0",
    "type": "object",
    "required": ["scheduler", "graphdatabase"],
    "properties": {
        "scheduler": {
            "description": "metadata for the rainbow scheduler",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "secret": {"type": "string"},
                "algorithm": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "options": {"type": "object"},
                    },
                },
                "user": {"type": "object"},
            },
            "additionalProperties": False,
        },
        "graphdatabase": {
            "description": "metadata for the rainbow graph database",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "options": {"type": "object"},
            },
            "additionalProperties": False,
        },
        "clusters": {
            "description": "listing of known clusters to submit to",
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "token": {"type": "string"},
                },
                "additionalProperties": False,
            },
            "additionalProperties": False,
        },
    },
}

# Note that this has experimental features added, they are flagged
jobspec_v1 = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "http://github.com/flux-framework/rfc/tree/master/data/spec_24/schema.json",
    "title": "jobspec-01",
    "description": "Flux jobspec version 1",
    "definitions": {
        "intranode_resource_vertex": {
            "description": "schema for resource vertices within a node, cannot have child vertices",
            "type": "object",
            "required": ["type", "count"],
            "properties": {
                "type": {"enum": ["core", "gpu"]},
                "count": {"type": "integer", "minimum": 1},
                "unit": {"type": "string"},
            },
            "additionalProperties": False,
        },
        "node_vertex": {
            "description": "schema for the node resource vertex",
            "type": "object",
            "required": ["type", "count", "with"],
            "properties": {
                "type": {"enum": ["node"]},
                "count": {"type": "integer", "minimum": 1},
                "unit": {"type": "string"},
                "with": {
                    "type": "array",
                    "minItems": 1,
                    "maxItems": 1,
                    "items": {"oneOf": [{"$ref": "#/definitions/slot_vertex"}]},
                },
            },
            "additionalProperties": False,
        },
        "slot_vertex": {
            "description": "special slot resource type - label assigns to task slot",
            "type": "object",
            "required": ["type", "count", "with", "label"],
            "properties": {
                "type": {"enum": ["slot"]},
                "count": {"type": "integer", "minimum": 1},
                "unit": {"type": "string"},
                "label": {"type": "string"},
                "exclusive": {"type": "boolean"},
                "with": {
                    "type": "array",
                    "minItems": 1,
                    "maxItems": 2,
                    "items": {"oneOf": [{"$ref": "#/definitions/intranode_resource_vertex"}]},
                },
            },
            "additionalProperties": False,
        },
    },
    "type": "object",
    # NOTE that I removed resources, I don't see why they need to be required
    "required": ["version", "resources", "tasks"],
    # "required": ["version", "resources", "attributes", "tasks"],
    "properties": {
        "version": {
            "description": "the jobspec version",
            "type": "integer",
            "enum": [1],
        },
        "resources": {
            "description": "requested resources",
            "type": "array",
            "minItems": 1,
            "maxItems": 1,
            "items": {
                "oneOf": [
                    {"$ref": "#/definitions/node_vertex"},
                    {"$ref": "#/definitions/slot_vertex"},
                ]
            },
        },
        "attributes": {
            "description": "system and user attributes",
            "type": ["object", "null"],
            "properties": {
                "system": {
                    "type": "object",
                    "properties": {
                        "duration": {"type": "number", "minimum": 0},
                        "cwd": {"type": "string"},
                        "environment": {"type": "object"},
                    },
                },
                "user": {"type": "object"},
            },
            "additionalProperties": False,
        },
        "tasks": {
            "description": "task configuration",
            "type": "array",
            "maxItems": 1,
            "items": {
                "type": "object",
                "required": ["command", "slot", "count"],
                "properties": {
                    "command": {
                        "type": ["string", "array"],
                        "minItems": 1,
                        "items": {"type": "string"},
                    },
                    # RESOURCES ARE EXPERIMENTAL
                    "resources": {"type": "object"},
                    "slot": {"type": "string"},
                    "count": {
                        "type": "object",
                        "additionalProperties": False,
                        "properties": {
                            "per_slot": {"type": "integer", "minimum": 1},
                            "total": {"type": "integer", "minimum": 1},
                        },
                    },
                },
                "additionalProperties": False,
            },
        },
    },
}



# Note that jsonschema used to be a part of flux core but was dropped
# https://github.com/flux-framework/flux-core/pull/5678/files

import json
import os

import jsonschema
import yaml

import rainbow.schema as schema
import rainbow.utils as utils


class JobspecV1:
    def __init__(self, filename, validate=True, schema=schema.jobspec_v1):
        """
        Load in and validate a Flux Jobspec
        """
        self.schema = schema
        self.jobspec = None
        self.load(filename)
        if validate:
            self.validate()

    def to_str(self):
        """
        Convert to string, which is needed for the submit.
        """
        return json.dumps(self.jobspec)

    def to_yaml(self):
        """
        Dump the jobspec to yaml string
        """
        return yaml.dump(self.jobspec)

    @property
    def name(self):
        try:
            return self.jobspec.get("tasks", {})[0]["command"][0]
        except Exception:
            return "app"

    def load(self, filename):
        """
        Load the jobspec
        """
        # Case 1: given a raw filename
        if isinstance(filename, str) and os.path.exists(filename):
            self.filename = os.path.abspath(filename)

            try:
                self.jobspec = utils.read_json(self.filename)
            except:
                self.jobspec = utils.read_yaml(self.filename)

        # Case 2: jobspec as dict (that we just want to validate)
        elif isinstance(filename, dict):
            self.jobspec = filename
        # Case 3: jobspec as string
        else:
            self.jobspec = json.loads(filename)

        # Case 4: wtf are you giving me? :X
        if not self.jobspec:
            raise ValueError("Unrecognized input format for jobspec.")

    def validate(self):
        """
        Validate the jsonschema
        """
        jsonschema.validate(self.jobspec, self.schema)



from .memory import MemoryBackend


def get_backend(name, options):
    """
    Get a named backend.
    """
    name = name.lower()
    if name == "memory":
        return MemoryBackend(**options)
    raise ValueError(f"Backend {name} is not a known graph database backend")



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class RegisterRequest(_message.Message):
    __slots__ = ("name", "secret", "nodes", "subsystem", "sent")
    NAME_FIELD_NUMBER: _ClassVar[int]
    SECRET_FIELD_NUMBER: _ClassVar[int]
    NODES_FIELD_NUMBER: _ClassVar[int]
    SUBSYSTEM_FIELD_NUMBER: _ClassVar[int]
    SENT_FIELD_NUMBER: _ClassVar[int]
    name: str
    secret: str
    nodes: str
    subsystem: str
    sent: _timestamp_pb2.Timestamp
    def __init__(self, name: _Optional[str] = ..., secret: _Optional[str] = ..., nodes: _Optional[str] = ..., subsystem: _Optional[str] = ..., sent: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class SubmitJobRequest(_message.Message):
    __slots__ = ("name", "clusters", "jobspec", "sent")
    class Cluster(_message.Message):
        __slots__ = ("name", "token")
        NAME_FIELD_NUMBER: _ClassVar[int]
        TOKEN_FIELD_NUMBER: _ClassVar[int]
        name: str
        token: str
        def __init__(self, name: _Optional[str] = ..., token: _Optional[str] = ...) -> None: ...
    NAME_FIELD_NUMBER: _ClassVar[int]
    CLUSTERS_FIELD_NUMBER: _ClassVar[int]
    JOBSPEC_FIELD_NUMBER: _ClassVar[int]
    SENT_FIELD_NUMBER: _ClassVar[int]
    name: str
    clusters: _containers.RepeatedCompositeFieldContainer[SubmitJobRequest.Cluster]
    jobspec: str
    sent: _timestamp_pb2.Timestamp
    def __init__(self, name: _Optional[str] = ..., clusters: _Optional[_Iterable[_Union[SubmitJobRequest.Cluster, _Mapping]]] = ..., jobspec: _Optional[str] = ..., sent: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class ReceiveJobsRequest(_message.Message):
    __slots__ = ("cluster", "secret", "maxJobs", "sent")
    CLUSTER_FIELD_NUMBER: _ClassVar[int]
    SECRET_FIELD_NUMBER: _ClassVar[int]
    MAXJOBS_FIELD_NUMBER: _ClassVar[int]
    SENT_FIELD_NUMBER: _ClassVar[int]
    cluster: str
    secret: str
    maxJobs: int
    sent: _timestamp_pb2.Timestamp
    def __init__(self, cluster: _Optional[str] = ..., secret: _Optional[str] = ..., maxJobs: _Optional[int] = ..., sent: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class AcceptJobsRequest(_message.Message):
    __slots__ = ("cluster", "secret", "jobids", "sent")
    CLUSTER_FIELD_NUMBER: _ClassVar[int]
    SECRET_FIELD_NUMBER: _ClassVar[int]
    JOBIDS_FIELD_NUMBER: _ClassVar[int]
    SENT_FIELD_NUMBER: _ClassVar[int]
    cluster: str
    secret: str
    jobids: _containers.RepeatedScalarFieldContainer[int]
    sent: _timestamp_pb2.Timestamp
    def __init__(self, cluster: _Optional[str] = ..., secret: _Optional[str] = ..., jobids: _Optional[_Iterable[int]] = ..., sent: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class RegisterResponse(_message.Message):
    __slots__ = ("request_id", "token", "secret", "status")
    class ResultType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        REGISTER_UNSPECIFIED: _ClassVar[RegisterResponse.ResultType]
        REGISTER_SUCCESS: _ClassVar[RegisterResponse.ResultType]
        REGISTER_ERROR: _ClassVar[RegisterResponse.ResultType]
        REGISTER_DENIED: _ClassVar[RegisterResponse.ResultType]
        REGISTER_EXISTS: _ClassVar[RegisterResponse.ResultType]
    REGISTER_UNSPECIFIED: RegisterResponse.ResultType
    REGISTER_SUCCESS: RegisterResponse.ResultType
    REGISTER_ERROR: RegisterResponse.ResultType
    REGISTER_DENIED: RegisterResponse.ResultType
    REGISTER_EXISTS: RegisterResponse.ResultType
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    SECRET_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    request_id: str
    token: str
    secret: str
    status: RegisterResponse.ResultType
    def __init__(self, request_id: _Optional[str] = ..., token: _Optional[str] = ..., secret: _Optional[str] = ..., status: _Optional[_Union[RegisterResponse.ResultType, str]] = ...) -> None: ...

class SubmitJobResponse(_message.Message):
    __slots__ = ("request_id", "jobid", "status")
    class ResultType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        SUBMIT_UNSPECIFIED: _ClassVar[SubmitJobResponse.ResultType]
        SUBMIT_SUCCESS: _ClassVar[SubmitJobResponse.ResultType]
        SUBMIT_ERROR: _ClassVar[SubmitJobResponse.ResultType]
        SUBMIT_DENIED: _ClassVar[SubmitJobResponse.ResultType]
    SUBMIT_UNSPECIFIED: SubmitJobResponse.ResultType
    SUBMIT_SUCCESS: SubmitJobResponse.ResultType
    SUBMIT_ERROR: SubmitJobResponse.ResultType
    SUBMIT_DENIED: SubmitJobResponse.ResultType
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    JOBID_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    request_id: str
    jobid: int
    status: SubmitJobResponse.ResultType
    def __init__(self, request_id: _Optional[str] = ..., jobid: _Optional[int] = ..., status: _Optional[_Union[SubmitJobResponse.ResultType, str]] = ...) -> None: ...

class ReceiveJobsResponse(_message.Message):
    __slots__ = ("request_id", "jobs", "status")
    class ResultType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        REQUEST_JOBS_NORESULTS: _ClassVar[ReceiveJobsResponse.ResultType]
        REQUEST_JOBS_SUCCESS: _ClassVar[ReceiveJobsResponse.ResultType]
        REQUEST_JOBS_ERROR: _ClassVar[ReceiveJobsResponse.ResultType]
        REQUEST_JOBS_DENIED: _ClassVar[ReceiveJobsResponse.ResultType]
    REQUEST_JOBS_NORESULTS: ReceiveJobsResponse.ResultType
    REQUEST_JOBS_SUCCESS: ReceiveJobsResponse.ResultType
    REQUEST_JOBS_ERROR: ReceiveJobsResponse.ResultType
    REQUEST_JOBS_DENIED: ReceiveJobsResponse.ResultType
    class JobsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: int
        value: str
        def __init__(self, key: _Optional[int] = ..., value: _Optional[str] = ...) -> None: ...
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    JOBS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    request_id: str
    jobs: _containers.ScalarMap[int, str]
    status: ReceiveJobsResponse.ResultType
    def __init__(self, request_id: _Optional[str] = ..., jobs: _Optional[_Mapping[int, str]] = ..., status: _Optional[_Union[ReceiveJobsResponse.ResultType, str]] = ...) -> None: ...

class AcceptJobsResponse(_message.Message):
    __slots__ = ("status",)
    class ResultType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        RESULT_TYPE_UNSPECIFIED: _ClassVar[AcceptJobsResponse.ResultType]
        RESULT_TYPE_PARTIAL: _ClassVar[AcceptJobsResponse.ResultType]
        RESULT_TYPE_SUCCESS: _ClassVar[AcceptJobsResponse.ResultType]
        RESULT_TYPE_ERROR: _ClassVar[AcceptJobsResponse.ResultType]
    RESULT_TYPE_UNSPECIFIED: AcceptJobsResponse.ResultType
    RESULT_TYPE_PARTIAL: AcceptJobsResponse.ResultType
    RESULT_TYPE_SUCCESS: AcceptJobsResponse.ResultType
    RESULT_TYPE_ERROR: AcceptJobsResponse.ResultType
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: AcceptJobsResponse.ResultType
    def __init__(self, status: _Optional[_Union[AcceptJobsResponse.ResultType, str]] = ...) -> None: ...
